@@ -1,38 +1,88 @@
-use crate::sparse_graph::SparseGraph;
+use crate::error::NautyError;
+use crate::nauty_graph::DenseGraph;
+#[cfg(feature = "libc")]
+use crate::nauty_graph::SparseGraph;
 
 use std::cmp::Ord;
+#[cfg(feature = "libc")]
+use std::convert::Infallible;
+use std::fmt::Debug;
 use std::hash::Hash;
 
 use nauty_Traces_sys::{
-    optionblk, sparsegraph, sparsenauty, statsblk, Traces, TracesOptions,
-    TracesStats, FALSE, SG_FREE, TRUE,
+    densenauty, empty_graph, optionblk, statsblk, FALSE, MTOOBIG, NTOOBIG, TRUE,
 };
+#[cfg(feature = "libc")]
+use nauty_Traces_sys::{
+    sparsegraph, sparsenauty, Traces, TracesOptions, TracesStats, SG_FREE,
+};
+#[cfg(feature = "libc")]
+use petgraph::graph::UnGraph;
 use petgraph::{
-    graph::{DiGraph, Graph, IndexType, UnGraph},
-    visit::EdgeRef,
+    graph::{Graph, IndexType},
     EdgeType,
 };
-use thiserror::Error;
 
 /// Find the canonical labelling for a graph
-///
-/// Internally, this uses Traces for undirected graphs without
-/// self-loops and sparse nauty otherwise.
 pub trait IntoCanon {
     fn into_canon(self) -> Self;
 }
 
+/// Try to find the canonical labelling for a graph
+pub trait TryIntoCanon {
+    type Error;
+
+    fn try_into_canon(self) -> Result<Self, Self::Error>
+    where
+        Self: Sized;
+}
+
 /// Use sparse nauty to find the canonical labelling
+///
+/// Requires the `libc` feature
+#[cfg(any(feature = "libc", doc))]
 pub trait IntoCanonNautySparse {
     fn into_canon_nauty_sparse(self) -> Self;
 }
 
+/// Use sparse nauty to find the canonical labelling
+///
+/// Requires the `libc` feature
+#[cfg(any(feature = "libc", doc))]
+pub trait TryIntoCanonNautySparse {
+    type Error;
+
+    fn try_into_canon_nauty_sparse(self) -> Result<Self, Self::Error>
+    where
+        Self: Sized;
+}
+
 /// Use dense nauty to find the canonical labelling
 pub trait IntoCanonNautyDense {
     fn into_canon_nauty_dense(self) -> Self;
 }
 
+/// Use dense nauty to find the canonical labelling
+pub trait TryIntoCanonNautyDense {
+    type Error;
+
+    fn try_into_canon_nauty_dense(self) -> Result<Self, Self::Error>
+    where
+        Self: Sized;
+}
+
 /// Use Traces to find the canonical labelling
+///
+/// Requires the `libc` feature
+#[cfg(any(feature = "libc", doc))]
+pub trait IntoCanonTraces {
+    fn into_canon_traces(self) -> Self;
+}
+
+/// Use Traces to find the canonical labelling
+///
+/// Requires the `libc` feature
+#[cfg(any(feature = "libc", doc))]
 pub trait TryIntoCanonTraces {
     type Error;
 
@@ -41,43 +91,45 @@ pub trait TryIntoCanonTraces {
         Self: Sized;
 }
 
-impl<N, E, Ix: IndexType> IntoCanon for UnGraph<N, E, Ix>
+impl<N, E, Ty: EdgeType, Ix: IndexType> IntoCanon for Graph<N, E, Ty, Ix>
+where
+    Graph<N, E, Ty, Ix>: TryIntoCanon,
+    <Graph<N, E, Ty, Ix> as TryIntoCanon>::Error: Debug,
+{
+    fn into_canon(self) -> Self {
+        self.try_into_canon().unwrap()
+    }
+}
+
+impl<N, E, Ty: EdgeType, Ix: IndexType> TryIntoCanon for Graph<N, E, Ty, Ix>
 where
     N: Ord,
     E: Hash + Ord,
 {
-    fn into_canon(self) -> Self {
-        match self.try_into_canon_traces() {
-            Ok(c) => c,
-            Err(TracesError::SelfLoop(g)) => g.into_canon_nauty_sparse(),
-        }
+    type Error = NautyError;
+
+    fn try_into_canon(self) -> Result<Self, Self::Error> {
+        self.try_into_canon_nauty_dense()
     }
 }
 
-impl<N, E, Ix: IndexType> IntoCanon for DiGraph<N, E, Ix>
-where
-    N: Ord,
-    E: Hash + Ord,
-{
-    fn into_canon(self) -> Self {
-        self.into_canon_nauty_sparse()
-    }
-}
-
-impl<N, E, Ty, Ix: IndexType> IntoCanonNautySparse for Graph<N, E, Ty, Ix>
+#[cfg(feature = "libc")]
+impl<N, E, Ty, Ix: IndexType> TryIntoCanonNautySparse for Graph<N, E, Ty, Ix>
 where
     N: Ord,
     E: Hash + Ord,
     Ty: EdgeType,
 {
-    fn into_canon_nauty_sparse(self) -> Self {
+    type Error = Infallible;
+
+    fn try_into_canon_nauty_sparse(self) -> Result<Self, Self::Error> {
         if self.node_count() == 0 {
-            return self;
+            return Ok(self);
         }
         let mut options = optionblk::default_sparse();
         options.getcanon = TRUE;
         options.defaultptn = FALSE;
-        options.digraph = TRUE;
+        options.digraph = if self.is_directed() { TRUE } else { FALSE };
         let mut stats = statsblk::default();
         let mut sg = SparseGraph::from(self);
         let mut orbits = vec![0; sg.g.v.len()];
@@ -94,28 +146,95 @@ where
             );
             SG_FREE(&mut cg);
         }
-        sg.into()
+        debug_assert_eq!(stats.errstatus, 0);
+        Ok(sg.into())
     }
 }
 
+#[cfg(feature = "libc")]
+impl<N, E, Ty, Ix> IntoCanonNautySparse for Graph<N, E, Ty, Ix>
+where
+    Graph<N, E, Ty, Ix>: TryIntoCanonNautySparse,
+    <Graph<N, E, Ty, Ix> as TryIntoCanonNautySparse>::Error: Debug,
+{
+    fn into_canon_nauty_sparse(self) -> Self {
+        self.try_into_canon_nauty_sparse().unwrap()
+    }
+}
+
+impl<N, E, Ty, Ix: IndexType> TryIntoCanonNautyDense for Graph<N, E, Ty, Ix>
+where
+    N: Ord,
+    E: Hash + Ord,
+    Ty: EdgeType,
+{
+    type Error = NautyError;
+
+    fn try_into_canon_nauty_dense(self) -> Result<Self, Self::Error> {
+        use ::std::os::raw::c_int;
+        use NautyError::*;
+
+        if self.node_count() == 0 {
+            return Ok(self);
+        }
+        let mut options = optionblk {
+            getcanon: TRUE,
+            defaultptn: FALSE,
+            digraph: if self.is_directed() { TRUE } else { FALSE },
+            ..Default::default()
+        };
+        let mut stats = statsblk::default();
+        let mut dg = DenseGraph::from(self);
+        let mut orbits = vec![0; dg.n];
+        let mut cg = empty_graph(dg.m, dg.n);
+        unsafe {
+            densenauty(
+                dg.g.as_mut_ptr(),
+                dg.nodes.lab.as_mut_ptr(),
+                dg.nodes.ptn.as_mut_ptr(),
+                orbits.as_mut_ptr(),
+                &mut options,
+                &mut stats,
+                dg.m as c_int,
+                dg.n as c_int,
+                cg.as_mut_ptr(),
+            );
+        }
+        match stats.errstatus {
+            0 => Ok(dg.into()),
+            MTOOBIG => Err(MTooBig),
+            NTOOBIG => Err(NTooBig),
+            _ => unreachable!(),
+        }
+    }
+}
+
+impl<N, E, Ty, Ix> IntoCanonNautyDense for Graph<N, E, Ty, Ix>
+where
+    Graph<N, E, Ty, Ix>: TryIntoCanonNautyDense,
+    <Graph<N, E, Ty, Ix> as TryIntoCanonNautyDense>::Error: Debug,
+{
+    fn into_canon_nauty_dense(self) -> Self {
+        self.try_into_canon_nauty_dense().unwrap()
+    }
+}
+
+#[cfg(feature = "libc")]
 impl<N, E, Ix: IndexType> TryIntoCanonTraces for UnGraph<N, E, Ix>
 where
     N: Ord,
     E: Hash + Ord,
 {
-    type Error = TracesError<N, E, Ix>;
+    type Error = Infallible;
 
     fn try_into_canon_traces(self) -> Result<Self, Self::Error> {
         if self.node_count() == 0 {
             return Ok(self);
         }
-        if has_self_loop(&self) {
-            return Err(TracesError::SelfLoop(self));
-        }
         let mut options = TracesOptions {
             getcanon: TRUE,
             defaultptn: FALSE,
-            digraph: TRUE,
+            digraph: FALSE,
             ..Default::default()
         };
         let mut stats = TracesStats::default();
@@ -134,34 +253,33 @@ where
             );
             SG_FREE(&mut cg);
         }
+        debug_assert_eq!(stats.errstatus, 0);
         Ok(sg.into())
     }
 }
 
-#[derive(Error, Debug)]
-pub enum TracesError<N, E, Ix: IndexType> {
-    #[error("Graph has at least one self-loop")]
-    SelfLoop(UnGraph<N, E, Ix>),
-}
-
-fn has_self_loop<N, E, Ty: EdgeType, Ix: IndexType>(
-    g: &Graph<N, E, Ty, Ix>,
-) -> bool {
-    g.edge_references().any(|e| e.source() == e.target())
+#[cfg(feature = "libc")]
+impl<N, E, Ix> IntoCanonTraces for UnGraph<N, E, Ix>
+where
+    UnGraph<N, E, Ix>: TryIntoCanonTraces,
+    <UnGraph<N, E, Ix> as TryIntoCanonTraces>::Error: Debug,
+{
+    fn into_canon_traces(self) -> Self {
+        self.try_into_canon_traces().unwrap()
+    }
 }
 
 #[cfg(test)]
 mod tests {
     use super::super::cmp::IsIdentical;
     use super::*;
-    use itertools::izip;
     use petgraph::{
         algo::isomorphism::is_isomorphic,
-        graph::{Graph, IndexType},
-        Directed, EdgeType, Undirected,
+        graph::{Graph, UnGraph},
+        Directed, Undirected,
     };
     use rand::prelude::*;
-    use testing::GraphIter;
+    use testing::{randomize_labels, GraphIter};
 
     use rand_xoshiro::Xoshiro256Plus;
 
@@ -203,41 +321,9 @@ mod tests {
         assert!(g1.is_identical(&g2));
     }
 
-    fn randomize_labels<N, E, Ty, Ix>(
-        g: Graph<N, E, Ty, Ix>,
-        rng: &mut impl Rng,
-    ) -> Graph<N, E, Ty, Ix>
-    where
-        Ty: EdgeType,
-        Ix: IndexType,
-    {
-        use petgraph::visit::NodeIndexable;
-        let mut res = Graph::with_capacity(g.node_count(), g.edge_count());
-        let edges = Vec::from_iter(g.edge_references().map(|e| {
-            let source = g.to_index(e.source());
-            let target = g.to_index(e.target());
-            (source, target)
-        }));
-        let (nodes, edge_wts) = g.into_nodes_edges();
-        let mut nodes =
-            Vec::from_iter(nodes.into_iter().map(|n| n.weight).enumerate());
-        nodes.shuffle(rng);
-        let mut relabel = vec![0; nodes.len()];
-        for (new, (old, w)) in nodes.into_iter().enumerate() {
-            res.add_node(w);
-            relabel[old] = new;
-        }
-        let edges = izip!(edges, edge_wts).map(|((source, target), w)| {
-            (relabel[source], relabel[target], w.weight)
-        });
-        for (source, target, w) in edges {
-            res.add_edge(res.from_index(source), res.from_index(target), w);
-        }
-        res
-    }
-
+    #[cfg(feature = "libc")]
     #[test]
-    fn random_canon_undirected() {
+    fn random_canon_nauty_sparse_undirected() {
         log_init();
 
         let mut rng = Xoshiro256Plus::seed_from_u64(0);
@@ -248,18 +334,19 @@ mod tests {
             let gg = randomize_labels(g.clone(), &mut rng);
             debug!("Randomised graph: {gg:#?}");
             assert!(is_isomorphic(&g, &gg));
-            let g = g.into_canon();
+            let g = g.into_canon_nauty_sparse();
             debug!("Canonical graph (from initial): {g:#?}");
             assert!(is_isomorphic(&g, &gg));
-            let gg = gg.into_canon();
+            let gg = gg.into_canon_nauty_sparse();
             debug!("Canonical graph (from randomised): {gg:#?}");
             assert!(is_isomorphic(&g, &gg));
             assert!(g.is_identical(&gg));
         }
     }
 
+    #[cfg(feature = "libc")]
     #[test]
-    fn random_canon_directed() {
+    fn random_canon_nauty_sparse_directed() {
         log_init();
 
         let mut rng = Xoshiro256Plus::seed_from_u64(0);
@@ -270,10 +357,77 @@ mod tests {
             let gg = randomize_labels(g.clone(), &mut rng);
             debug!("Randomised graph: {gg:#?}");
             assert!(is_isomorphic(&g, &gg));
-            let g = g.into_canon();
+            let g = g.into_canon_nauty_sparse();
             debug!("Canonical graph (from initial): {g:#?}");
             assert!(is_isomorphic(&g, &gg));
-            let gg = gg.into_canon();
+            let gg = gg.into_canon_nauty_sparse();
+            debug!("Canonical graph (from randomised): {gg:#?}");
+            assert!(is_isomorphic(&g, &gg));
+            assert!(g.is_identical(&gg));
+        }
+    }
+
+    #[test]
+    fn random_canon_nauty_dense_undirected() {
+        log_init();
+
+        let mut rng = Xoshiro256Plus::seed_from_u64(0);
+        let graphs = GraphIter::<Undirected>::default();
+
+        for g in graphs.take(1000) {
+            debug!("Initial graph: {g:#?}");
+            let gg = randomize_labels(g.clone(), &mut rng);
+            debug!("Randomised graph: {gg:#?}");
+            assert!(is_isomorphic(&g, &gg));
+            let g = g.into_canon_nauty_dense();
+            debug!("Canonical graph (from initial): {g:#?}");
+            assert!(is_isomorphic(&g, &gg));
+            let gg = gg.into_canon_nauty_dense();
+            debug!("Canonical graph (from randomised): {gg:#?}");
+            assert!(is_isomorphic(&g, &gg));
+            assert!(g.is_identical(&gg));
+        }
+    }
+
+    #[test]
+    fn random_canon_nauty_dense_directed() {
+        log_init();
+
+        let mut rng = Xoshiro256Plus::seed_from_u64(0);
+        let graphs = GraphIter::<Directed>::default();
+
+        for g in graphs.take(700) {
+            debug!("Initial graph: {g:#?}");
+            let gg = randomize_labels(g.clone(), &mut rng);
+            debug!("Randomised graph: {gg:#?}");
+            assert!(is_isomorphic(&g, &gg));
+            let g = g.into_canon_nauty_dense();
+            debug!("Canonical graph (from initial): {g:#?}");
+            assert!(is_isomorphic(&g, &gg));
+            let gg = gg.into_canon_nauty_dense();
+            debug!("Canonical graph (from randomised): {gg:#?}");
+            assert!(is_isomorphic(&g, &gg));
+            assert!(g.is_identical(&gg));
+        }
+    }
+
+    #[cfg(feature = "libc")]
+    #[test]
+    fn random_canon_traces_undirected() {
+        log_init();
+
+        let mut rng = Xoshiro256Plus::seed_from_u64(0);
+        let graphs = GraphIter::<Undirected>::default();
+
+        for g in graphs.take(1000) {
+            debug!("Initial graph: {g:#?}");
+            let gg = randomize_labels(g.clone(), &mut rng);
+            debug!("Randomised graph: {gg:#?}");
+            assert!(is_isomorphic(&g, &gg));
+            let g = g.into_canon_traces();
+            debug!("Canonical graph (from initial): {g:#?}");
+            assert!(is_isomorphic(&g, &gg));
+            let gg = gg.into_canon_traces();
             debug!("Canonical graph (from randomised): {gg:#?}");
             assert!(is_isomorphic(&g, &gg));
             assert!(g.is_identical(&gg));
