@@ -5,16 +5,8 @@
 //! labellings](https://en.wikipedia.org/wiki/Graph_canonization) for
 //! [petgraph](https://github.com/petgraph/petgraph) graphs.
 //!
-//! # Usage
-//!
 //! Version 2.6 or 2.7 of the nauty and Traces library has to be
-//! installed first and _linked to the same C library as this crate_.
-//!
-//! Afterwards, add this to your Cargo.toml:
-//! ```toml
-//! [dependencies]
-//! nauty-pet = "0.4"
-//! ```
+//! installed separately before installing this crate.
 //!
 //! # Example
 //!
@@ -31,17 +23,39 @@
 //! assert_eq!(automorphism_info.grpsize(), 2.);
 //!
 //! // The canonical forms are identical
-//! let c1 = g1.into_canon();
-//! let c2 = g2.into_canon();
-//! assert!(c1.is_identical(&c2))
+//! let c1 = g1.clone().into_canon();
+//! let c2 = g2.clone().into_canon();
+//! assert!(c1.is_identical(&c2));
+//!
+//! // Alternatively, we can use a dedicated `struct` for canonically
+//! // labelled graphs
+//! let c1 = CanonGraph::from(g1);
+//! let c2 = CanonGraph::from(g2);
+//! assert_eq!(c1, c2);
+//! ```
+//!
+//! # Features
+//!
+//! To enable the use of sparse graphs for canonisation, ensure that
+//! the nauty and Traces library is linked to the same C library as
+//! this crate.
+//!
+//! Afterwards, add this to your Cargo.toml:
+//! ```toml
+//! [dependencies]
+//! nauty-pet = { version = "0.5", features = ["libc"] }
 //! ```
 mod autom;
-mod canon;
+pub mod canon;
 mod cmp;
+pub mod error;
+pub mod graph;
+mod nauty_graph;
 pub mod prelude;
-mod sparse_graph;
 
-pub use canon::{IntoCanon, IntoCanonNautySparse, TryIntoCanonTraces};
+pub use canon::IntoCanon;
+#[cfg(feature = "libc")]
+pub use canon::{IntoCanonNautySparse, TryIntoCanonTraces};
 pub use cmp::IsIdentical;
 
 #[cfg(test)]
